@@ -70,7 +70,7 @@ def _inbytes(x):
 
 def _randombits(x):
     if x % 8 != 0:
-        raise ValueError("lenght must be a multiple of 8")
+        raise ValueError("length must be a multiple of 8")
     return os.urandom(_inbytes(x))
 
 
@@ -161,7 +161,8 @@ class _RawNone(_RawJWS):
         return ''
 
     def verify(self, key, payload, signature):
-        raise InvalidSignature('The "none" signature cannot be verified')
+        if key.key_type != 'oct' or key.get_op_key() != '':
+            raise InvalidSignature('The "none" signature cannot be verified')
 
 
 class _HS256(_RawHMAC, JWAAlgorithm):
@@ -248,6 +249,18 @@ class _ES256(_RawEC, JWAAlgorithm):
         super(_ES256, self).__init__('P-256', hashes.SHA256())
 
 
+class _ES256K(_RawEC, JWAAlgorithm):
+
+    name = "ES256K"
+    description = "ECDSA using secp256k1 curve and SHA-256"
+    keysize = 256
+    algorithm_usage_location = 'alg'
+    algorithm_use = 'sig'
+
+    def __init__(self):
+        super(_ES256K, self).__init__('secp256k1', hashes.SHA256())
+
+
 class _ES384(_RawEC, JWAAlgorithm):
 
     name = "ES384"
@@ -693,8 +706,12 @@ class _EcdhEs(_RawKeyMgmt, JWAAlgorithm):
     def _check_key(self, key):
         if not isinstance(key, JWK):
             raise ValueError('key is not a JWK object')
-        if key.key_type != 'EC':
-            raise InvalidJWEKeyType('EC', key.key_type)
+        if key.key_type not in ['EC', 'OKP']:
+            raise InvalidJWEKeyType('EC or OKP', key.key_type)
+        if key.key_type == 'OKP':
+            if key.key_curve not in ['X25519', 'X448']:
+                raise InvalidJWEKeyType('X25519 or X448',
+                                        key.key_curve)
 
     def _derive(self, privkey, pubkey, alg, bitsize, headers):
         # OtherInfo is defined in NIST SP 56A 5.8.1.2.1
@@ -718,7 +735,13 @@ class _EcdhEs(_RawKeyMgmt, JWAAlgorithm):
 
         # no SuppPrivInfo
 
-        shared_key = privkey.exchange(ec.ECDH(), pubkey)
+        # Shared Key generation
+        if isinstance(privkey, ec.EllipticCurvePrivateKey):
+            shared_key = privkey.exchange(ec.ECDH(), pubkey)
+        else:
+            # X25519/X448
+            shared_key = privkey.exchange(pubkey)
+
         ckdf = ConcatKDFHash(algorithm=hashes.SHA256(),
                              length=_inbytes(bitsize),
                              otherinfo=otherinfo,
@@ -802,6 +825,28 @@ class _EcdhEsAes256Kw(_EcdhEs):
     algorithm_use = 'kex'
 
 
+class _EdDsa(_RawJWS, JWAAlgorithm):
+
+    name = 'EdDSA'
+    description = 'EdDSA using Ed25519 or Ed448 algorithms'
+    algorithm_usage_location = 'alg'
+    algorithm_use = 'sig'
+    keysize = None
+
+    def sign(self, key, payload):
+
+        if key.key_curve in ['Ed25519', 'Ed448']:
+            skey = key.get_op_key('sign')
+            return skey.sign(payload)
+        raise NotImplementedError
+
+    def verify(self, key, payload, signature):
+        if key.key_curve in ['Ed25519', 'Ed448']:
+            pkey = key.get_op_key('verify')
+            return pkey.verify(signature, payload)
+        raise NotImplementedError
+
+
 class _RawJWE(object):
 
     def encrypt(self, k, a, m):
@@ -1009,6 +1054,7 @@ class JWA(object):
         'RS384': _RS384,
         'RS512': _RS512,
         'ES256': _ES256,
+        'ES256K': _ES256K,
         'ES384': _ES384,
         'ES512': _ES512,
         'PS256': _PS256,
@@ -1026,6 +1072,7 @@ class JWA(object):
         'ECDH-ES+A128KW': _EcdhEsAes128Kw,
         'ECDH-ES+A192KW': _EcdhEsAes192Kw,
         'ECDH-ES+A256KW': _EcdhEsAes256Kw,
+        'EdDSA': _EdDsa,
         'A128GCMKW': _A128GcmKw,
         'A192GCMKW': _A192GcmKw,
         'A256GCMKW': _A256GcmKw,
